
def get_cloud_default_index_settings():
    """
    Diverges from default in:
        - number_of_shards
    """
    return {
        "index_defaults": {
            "treat_urls_and_pointers_as_images": False,
            "model": "hf/all_datasets_v4_MiniLM-L6",
            "normalize_embeddings": True,
            "text_preprocessing": {
                "split_length": 2,
                "split_overlap": 0,
                "split_method": "sentence"
            },
            "image_preprocessing": {
                "patch_method": None
            }
        },
        "number_of_shards": 1,
        "number_of_replicas": 0,
    }



import base64
import os
from typing import Any, Dict, List, Optional, Union

from pydantic import error_wrappers
from requests.exceptions import RequestException
from typing_extensions import deprecated

from marqo.cloud_helpers import cloud_wait_for_index_status
from marqo.default_instance_mappings import DefaultInstanceMappings
from marqo.index import Index
from marqo.config import Config
from marqo.instance_mappings import InstanceMappings
from marqo.marqo_cloud_instance_mappings import MarqoCloudInstanceMappings
from marqo.models import BulkSearchBody, BulkSearchQuery
from marqo._httprequests import HttpRequests
from marqo import utils, enums
from marqo import errors
from marqo.marqo_logging import mq_logger
from marqo.errors import MarqoWebError
# we want to avoid name conflicts with marqo.version
from json import JSONDecodeError

# A dictionary to cache the marqo url and version for compatibility check


class Client:
    """
    A client for the marqo API

    A client instance is needed for every marqo API method to know the location of
    marqo and its permissions.
    """

    def __init__(
            self, url: Optional[str] = "http://localhost:8882",
            instance_mappings: Optional[InstanceMappings] = None,
            main_user: str = None, main_password: str = None,
            return_telemetry: bool = False,
            api_key: str = None
    ) -> None:
        """
        Parameters
        ----------
        url:
            The url to the Marqo API (ex: http://localhost:8882) If MARQO_CLOUD_URL environment variable is set, when
            matching url is passed, the client will use the Marqo Cloud instance mappings.
        instance_mappings:
            An instance of InstanceMappings that maps index names to urls
        api_key:
            The api key to use for authentication with the Marqo API
        """
        if url is not None and instance_mappings is not None:
            raise ValueError("Cannot specify both url and instance_mappings")

        is_marqo_cloud = False
        if url is not None:
            if url.lower().startswith(os.environ.get("MARQO_CLOUD_URL", "https://api.marqo.ai")):
                instance_mappings = MarqoCloudInstanceMappings(control_base_url=url, api_key=api_key)
                is_marqo_cloud = True
            else:
                instance_mappings = DefaultInstanceMappings(url, main_user, main_password)

        self.config = Config(
            instance_mappings=instance_mappings,
            is_marqo_cloud=is_marqo_cloud,
            use_telemetry=return_telemetry,
            api_key=api_key
        )
        self.http = HttpRequests(self.config)

    def create_index(
            self, index_name: str,
            treat_urls_and_pointers_as_images=False, model=None,
            normalize_embeddings=True,
            sentences_per_chunk=2,
            sentence_overlap=0,
            image_preprocessing_method=None,
            settings_dict=None,
            inference_node_type=None,
            storage_node_type=None,
            inference_node_count=1,
            storage_node_count=1,
            replicas_count=0,
    ) -> Dict[str, Any]:
        """Create the index. Please refer to the marqo cloud to see options for inference and storage node types.

        Args:
            index_name: name of the index.
            treat_urls_and_pointers_as_images:
            model:
            normalize_embeddings:
            sentences_per_chunk:
            sentence_overlap:
            image_preprocessing_method:
            settings_dict: if specified, overwrites all other setting
                parameters, and is passed directly as the index's
                index_settings
            inference_node_type:
            storage_node_type:
            inference_node_count;
            storage_node_count:
            replicas_count:
        Returns:
            Response body, containing information about index creation result
        """
        return Index.create(
            config=self.config, index_name=index_name,
            treat_urls_and_pointers_as_images=treat_urls_and_pointers_as_images,
            model=model, normalize_embeddings=normalize_embeddings,
            sentences_per_chunk=sentences_per_chunk, sentence_overlap=sentence_overlap,
            image_preprocessing_method=image_preprocessing_method,
            settings_dict=settings_dict, inference_node_type=inference_node_type, storage_node_type=storage_node_type,
            storage_node_count=storage_node_count, replicas_count=replicas_count,
            inference_node_count=inference_node_count,
        )

    def delete_index(self, index_name: str) -> Dict[str, Any]:
        """Deletes an index

        Args:
            index_name: name of the index

        Returns:
            response body about the result of the delete request
        """
        try:
            res = self.http.delete(path=f"indexes/{index_name}")
            cloud_wait_for_index_status(self.http, index_name, enums.IndexStatus.DELETED)
        except errors.MarqoWebError as e:
            return e.message

    def get_index(self, index_name: str) -> Index:
        """Get the index.
        This index should already exist.

        Args:
            index_name: name of the index

        Returns:
            An Index instance containing the information of the fetched index.

        Raises:
        """
        ix = Index(self.config, index_name)
        # verify it exists:
        self.http.get(path=f"indexes/{index_name}/stats", index_name=index_name)
        return ix

    def index(self, index_name: str) -> Index:
        """Create a local reference to an index identified by index_name,
        without doing an HTTP call.

        Calling this method doesn't create an index on the Marqo instance, but
         grants access to all the other methods in the Index class.

        Args:
            index_name: name of the index

        Returns:
            An Index instance.
        """
        if index_name is not None:
            return Index(self.config, index_name=index_name)
        raise Exception('The index UID should not be None')

    def get_indexes(self) -> Dict[str, List[Index]]:
        """Get all indexes.

        Returns:
        Indexes, a dictionary with the name of indexes.
        """
        response = self.http.get(path='indexes')
        response['results'] = [
            Index(
                config=self.config,
                index_name=index_info["index_name"],
            )
            for index_info in response["results"]
        ]
        return response

    def bulk_search(self, queries: List[Dict[str, Any]], device: Optional[str] = None) -> Dict[str, Any]:
        try:
            parsed_queries = [BulkSearchBody(**q) for q in queries]
        except error_wrappers.ValidationError as e:
            raise errors.InvalidArgError(f"some parameters in search query(s) are invalid. Errors are: {e.errors()}")

        translated_device_param = f"{f'?&device={utils.translate_device_string_for_url(device)}' if device is not None else ''}"
        return self.http.post(
            f"indexes/bulk/search{translated_device_param}",
            body=BulkSearchQuery(queries=parsed_queries).json()
        )

    @staticmethod
    def _base64url_encode(
            data: bytes
    ) -> str:
        return base64.urlsafe_b64encode(data).decode('utf-8').replace('=', '')

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use `mq.index(index_name).get_marqo()` instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def get_marqo(self):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("get_marqo")
        return self.http.get(path="")

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use `mq.index(index_name).health()` instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def health(self):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("health")
        try:
            return self.http.get(path="health")
        except (MarqoWebError, RequestException, TypeError, KeyError) as e:
            raise errors.BadRequestError("Marqo encountered an error trying to check the health of the Marqo instance. "
                                         "If you are trying to check the health on Marqo Cloud, please note that "
                                         "the `client.health()` API is not supported on Marqo Cloud and will be removed in "
                                         "Marqo 2.0.0. Please Use `client.index('your-index-name').health()` instead. "
                                         "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details.")

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use 'mq.index(index_name).eject_model() instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def eject_model(self, model_name: str, model_device: str):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("eject_model")
        return self.http.delete(path=f"models?model_name={model_name}&model_device={model_device}")

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use 'mq.index(index_name).get_loaded_models() instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def get_loaded_models(self):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("get_loaded_models")
        return self.http.get(path="models")

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use 'mq.index(index_name).get_cuda_info() instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def get_cuda_info(self):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("get_cuda_info")
        return self.http.get(path="device/cuda")

    @deprecated(
        "This method is deprecated and will be removed in Marqo 2.0.0. "
        "Please use 'mq.index(index_name).get_cpu_info() instead. "
        "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details."
    )
    def get_cpu_info(self):
        if self.config.is_marqo_cloud:
            self.raise_error_for_cloud("get_cpu_info")
        return self.http.get(path="device/cpu")

    @staticmethod
    def raise_error_for_cloud(function_name: str = None):
        raise errors.BadRequestError(
            f"The `mq.{function_name}()` API is not supported on Marqo Cloud. "
            f"Please Use `mq.index('your-index-name').{function_name}()` instead. "
            "Check `https://docs.marqo.ai/1.1.0/API-Reference/indexes/` for more details.")




import json
from requests import Response
from http import HTTPStatus


class MarqoError(Exception):
    """Generic class for Marqo error handling
    These will be caught and returned to the user as 5xx internal errors

    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'MarqoError. Error message: {self.message}'


class MarqoApiError(MarqoError):
    """Error sent by Marqo API"""

    def __init__(self, error: str, request: Response) -> None:
        self.status_code = request.status_code
        self.code = None
        self.link = None
        self.type = None

        if request.text:
            json_data = json.loads(request.text)
            self.message = json_data
            self.code = json_data.get('status')
            self.link = ''
            self.type = ''
            if 'error' in json_data and 'root_cause' in json_data["error"]\
                    and len(json_data.get('error').get('root_cause')) > 0:
                self.type = json_data.get('error').get('root_cause')[0].get('type')
        else:
            self.message = error
        super().__init__(self.message)

    def __str__(self) -> str:
        if self.code and self.link:
            return f'MarqoApiError. Error code: {self.code}. Error message: {self.message} Error documentation: {self.link} Error type: {self.type}'

        return f'MarqoApiError. {self.message}'


# MARQO WEB ERROR

class MarqoWebError(Exception):

    status_code: int = None
    error_type: str = None
    message: str = None
    code: str = None
    link: str = ""

    def __init__(self, message: str, status_code: int = None,
                 error_type: str = None, code: str = None,
                 link: str = None) -> None:
        self.message = message

        if self.status_code is None:
            self.status_code = status_code

        if self.error_type is None:
            self.error_type = error_type

        if self.code is None:
            self.code = code

        if self.link is None:
            self.link = link
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'MarqoWebError: {self.__class__.__name__} Error message: {self.message}\n' \
               f'status_code: {self.status_code}, type: {self.error_type}, ' \
               f'code: {self.code}, link: {self.link}'


# ---MARQO USER ERRORS---

# --


class __InvalidRequestError(MarqoWebError):
    """abstract error"""

    def __init__(self, message: str):
        self.message = message

    error_type = "invalid_request"


class UnsupportedOperationError(__InvalidRequestError):
    code = "unsupported_operation"


class IndexAlreadyExistsError(__InvalidRequestError):
    code = "index_already_exists"
    status_code = HTTPStatus.CONFLICT


class IndexNotFoundError(__InvalidRequestError):
    code = "index_not_found"
    status_code = HTTPStatus.NOT_FOUND


class InvalidIndexNameError(__InvalidRequestError):
    code = "invalid_index_name"
    status_code = HTTPStatus.BAD_REQUEST


class InvalidDocumentIdError(__InvalidRequestError):
    code = "invalid_document_id"
    status_code = HTTPStatus.BAD_REQUEST


class InvalidFieldNameError(__InvalidRequestError):
    code = "invalid_field_name"
    status_code = HTTPStatus.BAD_REQUEST


class InvalidArgError(__InvalidRequestError):
    code = "invalid_argument"
    status_code = HTTPStatus.BAD_REQUEST


class BadRequestError(__InvalidRequestError):
    code = "bad_request"
    status_code = HTTPStatus.BAD_REQUEST


class DocumentNotFoundError(__InvalidRequestError):
    code = "document_not_found"
    status_code = HTTPStatus.NOT_FOUND


class NonTensorIndexError(__InvalidRequestError):
    """Error trying to use a non-neural OpenSearch index like a neural one"""
    code = "document_not_found"
    status_code = HTTPStatus.NOT_FOUND

# ---MARQO INTERNAL ERROR---


class InternalError(MarqoWebError):
    error_type = "internal"
    code = "internal"
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR


class BackendCommunicationError(InternalError):
    """Error when connecting to Marqo"""
    code = "backend_communication_error"
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR

    def __init__(self, message: str,) -> None:
        self.message = f"Error communicating with Marqo: {message}"


class BackendTimeoutError(InternalError):
    """Error when Marqo operation takes longer than expected"""
    code = "backend_timeout_error"
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR

    def __init__(self, message: str,) -> None:
        self.message = f"Timeout error communicating with Marqo: {message}"


# NON HTTP ERRORS:

class MarqoCloudIndexNotReadyError(MarqoError):
    """Error when Marqo index is not ready"""
    code = "index_not_ready_cloud"
    status_code = HTTPStatus.NOT_FOUND

    def __init__(self, index_name: str,) -> None:
        self.message = f"The Python client could not resolve the endpoint for the index name {index_name}. " \
                       f"This could be due to the index is still in the process of being created," \
                       f" or that the client's cache has not yet been updated.\n" \
                       f"- Please try again in a couple of minutes, or you can query the index status" \
                       f" with mq.index({index_name}).get_status function to see when index is ready.\n" \
                       f"- If the problem persists, please contact marqo support at support@marqo.ai"


class MarqoCloudIndexNotFoundError(MarqoError):
    """Error when Marqo index is not ready"""
    code = "index_not_found_cloud"
    status_code = HTTPStatus.NOT_FOUND

    def __init__(self, index_name: str,) -> None:
        self.message = f"The index name {index_name} does not exist in the Marqo cloud or client's cache" \
                       f" has not yet been updated. Please check the index name and try again.\n" \
                       f"- If the problem persists, please contact marqo support at support@marqo.ai"



from abc import ABC, abstractmethod

from typing import Dict, Optional


class InstanceMappings(ABC):
    """
    Abstract class for mapping index names to URLs.

    Implementations of this class must ensure at least asymptotically constant computational complexity for all methods.
    An inefficient implementation of this class can cause the Marqo client to be slow.

    The namespace of index names is with respect to a Client instance.
    Index names must be unique across all Marqo URLs referenced in the Mappings.
    If you want to support the same index name on different Marqo URLs, please use separate Client objects.
    """
    @abstractmethod
    def get_index_base_url(self, index_name: str) -> str:
        """
        Return the base URL for the given index.

        Args:
            index_name: The index name
        """
        pass

    @abstractmethod
    def get_control_base_url(self) -> str:
        """
        Return the base URL for index control operations such as index creation and deletion.
        """
        pass

    @abstractmethod
    def is_remote(self):
        pass



from typing import Dict, List, Optional, Union
from pydantic import BaseModel


class BaseMarqoModel(BaseModel):
    class Config:
        extra: str = "forbid"
    pass


class SearchBody(BaseMarqoModel):
    q: Union[str, Dict[str, float]]
    searchableAttributes: Union[None, List[str]] = None
    searchMethod: Union[None, str] = "TENSOR"
    limit: int = 10
    offset: int = 0
    showHighlights: bool = True
    reRanker: str = None
    filter: str = None
    attributesToRetrieve: Union[None, List[str]] = None
    boost: Optional[Dict] = None
    image_download_headers: Optional[Dict] = None
    context: Optional[Dict] = None
    scoreModifiers: Optional[Dict] = None
    modelAuth: Optional[Dict] = None

class BulkSearchBody(SearchBody):
    index: str


class BulkSearchQuery(BaseMarqoModel):
    queries: List[BulkSearchBody]



import time
from typing import Optional

import requests
from requests.exceptions import Timeout

from marqo.errors import (
    MarqoCloudIndexNotFoundError,
    MarqoCloudIndexNotReadyError,
)
from marqo.instance_mappings import InstanceMappings
from marqo.marqo_logging import mq_logger


class MarqoCloudInstanceMappings(InstanceMappings):
    def __init__(self, control_base_url, api_key=None, url_cache_duration: int = 15):
        self.latest_index_mappings_refresh_timestamp = time.time() - url_cache_duration
        self._urls_mapping = {"READY": {}, "CREATING": {}}
        self.api_key = api_key
        self.url_cache_duration = url_cache_duration
        self._control_base_url = control_base_url

    def get_control_base_url(self) -> str:
        return f"{self._control_base_url}/api"

    def get_index_base_url(self, index_name: str) -> str:
        self._refresh_urls_if_needed(index_name)
        if index_name in self._urls_mapping['READY']:
            return self._urls_mapping['READY'][index_name]
        if index_name in self._urls_mapping['CREATING']:
            raise MarqoCloudIndexNotReadyError(index_name)
        raise MarqoCloudIndexNotFoundError(index_name)

    def is_remote(self):
        return True

    def _refresh_urls_if_needed(self, index_name):
        if index_name not in self._urls_mapping['READY'] and \
                time.time() - self.latest_index_mappings_refresh_timestamp > self.url_cache_duration:
            # fast refresh to catch if index was created
            self._refresh_urls()
        if index_name in self._urls_mapping['READY'] and \
                time.time() - self.latest_index_mappings_refresh_timestamp > 360:
            # slow refresh in case index was deleted
            self._refresh_urls(timeout=3)

    def _refresh_urls(self, timeout=None):
        try:
            response = requests.get(f'{self.get_control_base_url()}/indexes',
                                    headers={"x-api-key": self.api_key}, timeout=timeout)
        except Timeout:
            mq_logger.warning(
                f"Timeout getting and caching URLs for Marqo Cloud indexes from the"
                f" /api/indexes/ endpoint. Please contact marqo support at support@marqo.ai if this message"
                f" persists."
            )
            return None

        if not response.ok:
            mq_logger.warning(response.text)
        response_json = response.json()
        for index in response_json['results']:
            if index.get('index_status') in ["READY", "CREATING"]:
                self._urls_mapping[index['index_status']][index['index_name']] = index.get('endpoint')
        if self._urls_mapping:
            self.latest_index_mappings_refresh_timestamp = time.time()
